@@ -36,11 +36,13 @@
 #include <haproxy/obj_type-t.h>
 #include <haproxy/port_range-t.h>
 #include <haproxy/protocol-t.h>
+#include <haproxy/show_flags-t.h>
 #include <haproxy/thread-t.h>
 
 /* referenced below */
 struct connection;
-struct conn_stream;
+struct stconn;
+struct sedesc;
 struct cs_info;
 struct buffer;
 struct proxy;
@@ -50,6 +52,7 @@ struct pipe;
 struct quic_conn;
 struct bind_conf;
 struct qcs;
+struct ssl_sock_ctx;
 
 /* Note: subscribing to these events is only valid after the caller has really
  * attempted to perform the operation, and failed to proceed or complete.
@@ -59,50 +62,6 @@ enum sub_event_type {
 	SUB_RETRY_SEND       = 0x00000002,  /* Schedule the tasklet when we can attempt to send again */
 };
 
-/* conn_stream flags */
-enum {
-	CS_FL_NONE          = 0x00000000,  /* Just for initialization purposes */
-	CS_FL_SHRD          = 0x00000010,  /* read shut, draining extra data */
-	CS_FL_SHRR          = 0x00000020,  /* read shut, resetting extra data */
-	CS_FL_SHR           = CS_FL_SHRD | CS_FL_SHRR, /* read shut status */
-
-	CS_FL_SHWN          = 0x00000040,  /* write shut, verbose mode */
-	CS_FL_SHWS          = 0x00000080,  /* write shut, silent mode */
-	CS_FL_SHW           = CS_FL_SHWN | CS_FL_SHWS, /* write shut status */
-
-
-	CS_FL_ERROR         = 0x00000100,  /* a fatal error was reported */
-	CS_FL_RCV_MORE      = 0x00000200,  /* We may have more bytes to transfer */
-	CS_FL_WANT_ROOM     = 0x00000400,  /* More bytes to transfer, but not enough room */
-	CS_FL_ERR_PENDING   = 0x00000800,  /* An error is pending, but there's still data to be read */
-	CS_FL_EOS           = 0x00001000,  /* End of stream delivered to data layer */
-	/* unused: 0x00002000 */
-	CS_FL_EOI           = 0x00004000,  /* end-of-input reached */
-	CS_FL_MAY_SPLICE    = 0x00008000,  /* caller may use rcv_pipe() only if this flag is set */
-	CS_FL_WAIT_FOR_HS   = 0x00010000,  /* This stream is waiting for handhskae */
-	CS_FL_KILL_CONN     = 0x00020000,  /* must kill the connection when the CS closes */
-
-	/* following flags are supposed to be set by the mux and read/unset by
-	 * the stream-interface :
-	 */
-	CS_FL_NOT_FIRST     = 0x00100000,  /* this stream is not the first one */
-
-	/* flags set by the mux relayed to the stream */
-	CS_FL_WEBSOCKET     = 0x00200000,  /* websocket stream */
-};
-
-/* cs_shutr() modes */
-enum cs_shr_mode {
-	CS_SHR_DRAIN        = 0,           /* read shutdown, drain any extra stuff */
-	CS_SHR_RESET        = 1,           /* read shutdown, reset any extra stuff */
-};
-
-/* cs_shutw() modes */
-enum cs_shw_mode {
-	CS_SHW_NORMAL       = 0,           /* regular write shutdown */
-	CS_SHW_SILENT       = 1,           /* imminent close, don't notify peer */
-};
-
 /* For each direction, we have a CO_FL_XPRT_<DIR>_ENA flag, which
  * indicates if read or write is desired in that direction for the respective
  * layers. The current status corresponding to the current layer being used is
@@ -115,7 +74,9 @@ enum cs_shw_mode {
  * conn_cond_update_polling().
  */
 
-/* flags for use in connection->flags */
+/* flags for use in connection->flags. Please also update the conn_show_flags()
+ * function below in case of changes.
+ */
 enum {
 	CO_FL_NONE          = 0x00000000,  /* Just for initialization purposes */
 
@@ -143,8 +104,8 @@ enum {
 	CO_FL_WAIT_ROOM     = 0x00000800,  /* data sink is full */
 
 	/* These flags are used to report whether the from/to addresses are set or not */
-	CO_FL_ADDR_FROM_SET = 0x00001000,  /* addr.from is set */
-	CO_FL_ADDR_TO_SET   = 0x00002000,  /* addr.to is set */
+	/* unused: 0x00001000 */
+	/* unused: 0x00002000 */
 
 	CO_FL_EARLY_SSL_HS  = 0x00004000,  /* We have early data pending, don't start SSL handshake yet */
 	CO_FL_EARLY_DATA    = 0x00008000,  /* At least some of the data are early data */
@@ -159,6 +120,8 @@ enum {
 	CO_FL_ERROR         = 0x00100000,  /* a fatal error was reported     */
 	CO_FL_NOTIFY_DONE   = 0x001C0000,  /* any xprt shut/error flags above needs to be reported */
 
+	CO_FL_FDLESS        = 0x00200000,  /* this connection doesn't use any FD (e.g. QUIC) */
+
 	/* flags used to report connection status updates */
 	CO_FL_WAIT_L4_CONN  = 0x00400000,  /* waiting for L4 to be connected */
 	CO_FL_WAIT_L6_CONN  = 0x00800000,  /* waiting for L6 to be connected (eg: SSL) */
@@ -197,6 +160,29 @@ enum {
 	CO_FL_SOCKS4        = CO_FL_SOCKS4_SEND | CO_FL_SOCKS4_RECV,
 };
 
+/* This function is used to report flags in debugging tools. Please reflect
+ * below any single-bit flag addition above in the same order via the
+ * __APPEND_FLAG macro. The new end of the buffer is returned.
+ */
+static forceinline char *conn_show_flags(char *buf, size_t len, const char *delim, uint flg)
+{
+#define _(f, ...) __APPEND_FLAG(buf, len, delim, flg, f, #f, __VA_ARGS__)
+	/* prologue */
+	_(0);
+	/* flags */
+	_(CO_FL_SAFE_LIST, _(CO_FL_IDLE_LIST, _(CO_FL_CTRL_READY, _(CO_FL_XPRT_READY,
+	_(CO_FL_WANT_DRAIN, _(CO_FL_WAIT_ROOM, _(CO_FL_EARLY_SSL_HS, _(CO_FL_EARLY_DATA,
+	_(CO_FL_SOCKS4_SEND, _(CO_FL_SOCKS4_RECV, _(CO_FL_SOCK_RD_SH, _(CO_FL_SOCK_WR_SH,
+	_(CO_FL_ERROR, _(CO_FL_FDLESS, _(CO_FL_WAIT_L4_CONN, _(CO_FL_WAIT_L6_CONN,
+	_(CO_FL_SEND_PROXY, _(CO_FL_ACCEPT_PROXY, _(CO_FL_ACCEPT_CIP, _(CO_FL_SSL_WAIT_HS,
+	_(CO_FL_PRIVATE, _(CO_FL_RCVD_PROXY, _(CO_FL_SESS_IDLE, _(CO_FL_XPRT_TRACKED
+	))))))))))))))))))))))));
+	/* epilogue */
+	_(~0U);
+	return buf;
+#undef _
+}
+
 /* Possible connection error codes.
  * Warning: Do not reorder the codes, they are fetchable through the
  * "fc_err" sample fetch. If a new code is added, please add an error label
@@ -293,6 +279,18 @@ enum {
 	CO_SFL_STREAMER    = 0x0002,    /* Producer is continuously streaming data */
 };
 
+/* mux->shutr() modes */
+enum co_shr_mode {
+	CO_SHR_DRAIN        = 0,           /* read shutdown, drain any extra stuff */
+	CO_SHR_RESET        = 1,           /* read shutdown, reset any extra stuff */
+};
+
+/* mux->shutw() modes */
+enum co_shw_mode {
+	CO_SHW_NORMAL       = 0,           /* regular write shutdown */
+	CO_SHW_SILENT       = 1,           /* imminent close, don't notify peer */
+};
+
 /* known transport layers (for ease of lookup) */
 enum {
 	XPRT_RAW = 0,
@@ -305,10 +303,10 @@ enum {
 /* MUX-specific flags */
 enum {
 	MX_FL_NONE        = 0x00000000,
-	MX_FL_CLEAN_ABRT  = 0x00000001, /* abort is clearly reported as an error */
-	MX_FL_HTX         = 0x00000002, /* set if it is an HTX multiplexer */
-	MX_FL_HOL_RISK    = 0x00000004, /* set if the protocol is subject the to head-of-line blocking on server */
-	MX_FL_NO_UPG      = 0x00000008, /* set if mux does not support any upgrade */
+	MX_FL_HTX         = 0x00000001, /* set if it is an HTX multiplexer */
+	MX_FL_HOL_RISK    = 0x00000002, /* set if the protocol is subject the to head-of-line blocking on server */
+	MX_FL_NO_UPG      = 0x00000004, /* set if mux does not support any upgrade */
+	MX_FL_FRAMED      = 0x00000008, /* mux working on top of a framed transport layer (QUIC) */
 };
 
 /* PROTO token registration */
@@ -367,10 +365,12 @@ struct wait_event {
 };
 
 /* A connection handle is how we differentiate two connections on the lower
- * layers. It usually is a file descriptor but can be a connection id.
+ * layers. It usually is a file descriptor but can be a connection id. The
+ * CO_FL_FDLESS flag indicates which one is relevant.
  */
 union conn_handle {
-	int fd;                 /* file descriptor, for regular sockets */
+	struct quic_conn *qc;   /* Only present if this connection is a QUIC one (CO_FL_FDLESS=1) */
+	int fd;                 /* file descriptor, for regular sockets (CO_FL_FDLESS=0) */
 };
 
 /* xprt_ops describes transport-layer operations for a connection. They
@@ -401,6 +401,7 @@ struct xprt_ops {
 	int (*unsubscribe)(struct connection *conn, void *xprt_ctx, int event_type, struct wait_event *es); /* Unsubscribe <es> from events */
 	int (*remove_xprt)(struct connection *conn, void *xprt_ctx, void *toremove_ctx, const struct xprt_ops *newops, void *newctx); /* Remove an xprt from the connection, used by temporary xprt such as the handshake one */
 	int (*add_xprt)(struct connection *conn, void *xprt_ctx, void *toadd_ctx, const struct xprt_ops *toadd_ops, void **oldxprt_ctx, const struct xprt_ops **oldxprt_ops); /* Add a new XPRT as the new xprt, and return the old one */
+	struct ssl_sock_ctx *(*get_ssl_sock_ctx)(struct connection *); /* retrieve the ssl_sock_ctx in use, or NULL if none */
 	int (*show_fd)(struct buffer *, const struct connection *, const void *ctx); /* append some data about xprt for "show fd"; returns non-zero if suspicious */
 };
 
@@ -414,23 +415,20 @@ struct xprt_ops {
 struct mux_ops {
 	int  (*init)(struct connection *conn, struct proxy *prx, struct session *sess, struct buffer *input);  /* early initialization */
 	int  (*wake)(struct connection *conn);        /* mux-layer callback to report activity, mandatory */
-	size_t (*rcv_buf)(struct conn_stream *cs, struct buffer *buf, size_t count, int flags); /* Called from the upper layer to get data */
-	size_t (*snd_buf)(struct conn_stream *cs, struct buffer *buf, size_t count, int flags); /* Called from the upper layer to send data */
-	int  (*rcv_pipe)(struct conn_stream *cs, struct pipe *pipe, unsigned int count); /* recv-to-pipe callback */
-	int  (*snd_pipe)(struct conn_stream *cs, struct pipe *pipe); /* send-to-pipe callback */
-	void (*shutr)(struct conn_stream *cs, enum cs_shr_mode);     /* shutr function */
-	void (*shutw)(struct conn_stream *cs, enum cs_shw_mode);     /* shutw function */
+	size_t (*rcv_buf)(struct stconn *sc, struct buffer *buf, size_t count, int flags); /* Called from the upper layer to get data */
+	size_t (*snd_buf)(struct stconn *sc, struct buffer *buf, size_t count, int flags); /* Called from the upper layer to send data */
+	int  (*rcv_pipe)(struct stconn *sc, struct pipe *pipe, unsigned int count); /* recv-to-pipe callback */
+	int  (*snd_pipe)(struct stconn *sc, struct pipe *pipe); /* send-to-pipe callback */
+	void (*shutr)(struct stconn *sc, enum co_shr_mode);     /* shutr function */
+	void (*shutw)(struct stconn *sc, enum co_shw_mode);     /* shutw function */
 
-	struct conn_stream *(*attach)(struct connection *, struct session *sess); /* Create and attach a conn_stream to an outgoing connection */
-	const struct conn_stream *(*get_first_cs)(const struct connection *); /* retrieves any valid conn_stream from this connection */
-	void (*detach)(struct conn_stream *); /* Detach a conn_stream from an outgoing connection, when the request is done */
+	int (*attach)(struct connection *conn, struct sedesc *, struct session *sess); /* attach a stconn to an outgoing connection */
+	struct stconn *(*get_first_sc)(const struct connection *); /* retrieves any valid stconn from this connection */
+	void (*detach)(struct sedesc *); /* Detach an stconn from the stdesc from an outgoing connection, when the request is done */
 	int (*show_fd)(struct buffer *, struct connection *); /* append some data about connection into chunk for "show fd"; returns non-zero if suspicious */
-	int (*subscribe)(struct conn_stream *cs, int event_type,  struct wait_event *es); /* Subscribe <es> to events, such as "being able to send" */
-	int (*unsubscribe)(struct conn_stream *cs, int event_type,  struct wait_event *es); /* Unsubscribe <es> from events */
-	int (*ruqs_subscribe)(struct qcs *qcs, int event_type,  struct wait_event *es); /* Subscribe <es> to events, "being able to receive" only */
-	int (*ruqs_unsubscribe)(struct qcs *qcs, int event_type,  struct wait_event *es); /* Unsubscribe <es> from events */
-	int (*luqs_subscribe)(struct qcs *qcs, int event_type,  struct wait_event *es); /* Subscribe <es> to events, "being able to send" only */
-	int (*luqs_unsubscribe)(struct qcs *qcs, int event_type,  struct wait_event *es); /* Unsubscribe <es> from events */
+	int (*show_sd)(struct buffer *, struct sedesc *, const char *pfx); /* append some data about the mux stream into chunk for "show sess"; returns non-zero if suspicious */
+	int (*subscribe)(struct stconn *sc, int event_type,  struct wait_event *es); /* Subscribe <es> to events, such as "being able to send" */
+	int (*unsubscribe)(struct stconn *sc, int event_type,  struct wait_event *es); /* Unsubscribe <es> from events */
 	int (*avail_streams)(struct connection *conn); /* Returns the number of streams still available for a connection */
 	int (*avail_streams_bidi)(struct connection *conn); /* Returns the number of bidirectional streams still available for a connection */
 	int (*avail_streams_uni)(struct connection *conn); /* Returns the number of unidirectional streams still available for a connection */
@@ -450,18 +448,6 @@ struct mux_stopping_data {
 	struct task *task; /* task woken up on soft-stop */
 };
 
-/* data_cb describes the data layer's recv and send callbacks which are called
- * when I/O activity was detected after the transport layer is ready. These
- * callbacks are supposed to make use of the xprt_ops above to exchange data
- * from/to buffers and pipes. The <wake> callback is used to report activity
- * at the transport layer, which can be a connection opening/close, or any
- * data movement. It may abort a connection by returning < 0.
- */
-struct data_cb {
-	int  (*wake)(struct conn_stream *cs);  /* data-layer callback to report activity */
-	char name[8];                           /* data layer name, zero-terminated */
-};
-
 struct my_tcphdr {
 	uint16_t source;
 	uint16_t dest;
@@ -484,19 +470,6 @@ struct conn_src {
 #endif
 };
 
-/*
- * This structure describes the elements of a connection relevant to a stream
- */
-struct conn_stream {
-	enum obj_type obj_type;              /* differentiates connection from applet context */
-	/* 3 bytes hole here */
-	unsigned int flags;                  /* CS_FL_* */
-	struct connection *conn;             /* xprt-level connection */
-	void *data;                          /* pointer to upper layer's entity (eg: stream interface) */
-	const struct data_cb *data_cb;       /* data layer callbacks. Must be set before xprt->init() */
-	void *ctx;                           /* mux-specific context */
-};
-
 /* Hash header flag reflecting the input parameters present
  * CAUTION! Always update CONN_HASH_PARAMS_TYPE_COUNT when adding a new entry.
  */
@@ -565,7 +538,6 @@ struct connection {
 	struct sockaddr_storage *dst; /* destination address (pool), when known, otherwise NULL */
 	struct ist proxy_authority;   /* Value of the authority TLV received via PROXYv2 */
 	struct ist proxy_unique_id;   /* Value of the unique ID TLV received via PROXYv2 */
-	struct quic_conn *qc;         /* Only present if this connection is a QUIC one */
 
 	/* used to identify a backend connection for http-reuse,
 	 * thus only present if conn.target is of type OBJ_TYPE_SERVER
